﻿using AspectCore.DynamicProxy;
using Microsoft.Extensions.DependencyInjection;
using Microsoft.Extensions.Logging;
using System.Reflection;
using HealthMonitor.Core.Cache;
using HealthMonitor.Util.Models;

namespace HealthMonitor.Core.Pipeline.Aop.Query
{
    public class GetOrInsertEntityCacheValue : AopValueBase
    {
        private readonly static MethodInfo M_TaskFromResult = typeof(Task).GetMethod(nameof(Task.FromResult))!;

        public override async Task Invoke(CacheInterceptorContext context)
        {
            var entityType = context.EntityType;
            var returnType = context.ReturnType;
            var isAsync = context.IsAsyncMethod;
            var isEnumerable = context.IsReturnEnumerable;
            var param = context.AopContext.Parameters[0];
            string? key = null;

            //读取缓存
            var durableManager = context.AopContext.ServiceProvider.GetService<IDurableEntityManager>();
            var cacheHandler = durableManager?.GetCacheHandler(entityType);
            //若实体类型不支持缓存，则退出
            if (cacheHandler == null || cacheHandler.DurableSecond <= 0)
            {
                await InvokeNextAsync(context);
                return;
            }

            if (param is string) key = durableManager?.CalcCacheKey(entityType, param.ToString()!);
            else if (param is GeneralParam) key = durableManager?.CalcCacheKey(entityType, (param as GeneralParam)!);

            if (string.IsNullOrEmpty(key))
            {
                await InvokeNextAsync(context);
                return;
            }

            var cache = cacheHandler.GetEntitiesCache(key);
            object returnValue;
            if (cache != null)
            {
                context.Logger.LogInformation($"击中实体缓存{key}");

                returnValue = isEnumerable ? cache : cache.FirstOrDefault()!;
                if (isAsync) //&& methodReturnType == typeof(Task<>).MakeGenericType(returnType))
                {
                    //反射获取Task<>类型的返回值，相当于Task.FromResult(value);
                    context.AopContext.ReturnValue = M_TaskFromResult.MakeGenericMethod(returnType).Invoke(null, new[] { returnValue });
                    return;
                }

                context.AopContext.ReturnValue = returnValue;
                return;
            }

            await InvokeNextAsync(context);

            if (isAsync)
            {
                returnValue = await context.AopContext.UnwrapAsyncReturnValue();
                //反射获取异步结果的值，相当于(context.ReturnValue as Task<>).Result
                //returnValue = typeof(Task<>).MakeGenericType(returnType).GetProperty(nameof(Task<object>.Result)).GetValue(context.ReturnValue);
            }
            else
            {
                returnValue = context.AopContext.ReturnValue;
            }
            cacheHandler.SetEntityCache(key, returnValue, isEnumerable);
        }
    }
}
